import logging
import random
import time
from queue import Queue
from threading import Thread
from typing import Optional

from soco import SoCo

from socomote.action import *
from socomote.input_handler import InputHandler
from socomote.settings import GROUP_CONFIG, ZONES
from socomote.station import Station, Stations, is_station_uri
from socomote.tts_server import TTSServer

logger = logging.getLogger(__name__)

class Receiver:

    def __init__(self, controller: SoCo):
        self._controller: SoCo = controller
        self._handler = InputHandler()
        self._stations = Stations()  # todo - stations need refreshing
        self._tts_server = TTSServer()
        self._action_queue = Queue()
        self._action_thread = Thread(target=self._process_actions, daemon=True)

    def __enter__(self):
        self._tts_server.__enter__()
        self._action_thread.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._tts_server.__exit__(exc_type, exc_val, exc_tb)

    def start(self):
        for action in self._handler.actions():
            if self._action_queue.qsize() <= 3:
                self._action_queue.put(action)
            else:
                logger.info(f"Could not add {action} to action queue, queue full. Discarding action.")

    def _process_actions(self):
        while True:
            action = self._action_queue.get()
            logger.info(f"Performing action {action}")
            try:
                if isinstance(action, PlayPause):
                    self.play_pause()
                elif isinstance(action, VolUp):
                    self._controller.volume += 3
                elif isinstance(action, VolDown):
                    self._controller.volume -= 3
                elif isinstance(action, Next):
                    self.prev_next(is_next=True)
                elif isinstance(action, Previous):
                    self.prev_next(is_next=False)
                elif isinstance(action, SelectStation):
                    self.play_station_by_index(action.ix)
                elif isinstance(action, ShuffleStation):
                    station = random.choice(self._stations)
                    logger.info(f"Playing random station: {station}")
                    self.play_station(station)
                elif isinstance(action, Query):
                    self.query_station()
                elif isinstance(action, SelectGroup):
                    self.to_group(action.ix)
            except BaseException as e:
                logging.error(f"Unhandled exception: {e}")

    def _is_playing(self):
        # todo - get rid
        curr_info = self._controller.get_current_track_info()
        position = curr_info['position']
        logger.debug(f"Current position is {repr(position)}")
        if position is None or position == '' or position == '0:00:00':
            logger.debug(f"Currently paused.")
            return False
        else:
            logger.debug(f"Currently playing.")
            return True

    def _take_control(self):
        if self._controller.group.coordinator != self._controller:
            slaves = [s for s in self._controller.group if s != self._controller]
            logger.info(
                f"Controller {self._controller.player_name} is not the master, "
                f"taking control of: {[s.player_name for s in slaves]}"
            )
            for s in slaves:
                s.unjoin()
            for s in slaves:
                s.join(self._controller)

    def _play_uri(self, uri="", meta="", title="", start=True, force_radio=False):
        self._take_control()
        self._controller.play_uri(uri=uri, meta=meta, title=title, start=start, force_radio=force_radio)

    def _speak(self, text):
        uri = self._tts_server.get_uri(text)
        self._play_uri(uri=uri, title=text)

    def play_station(self, station: Station):
        logger.info(f"Playing station {station}")
        self._speak(station.title)
        time.sleep(2)  # todo
        start = time.time()
        self._play_uri(uri=station.uri, title=station.title)
        end = time.time()
        duration = end - start
        logger.debug(f"Took {duration} to play URI")

    def play_station_by_index(self, ix: int):
        station = self._stations[ix]
        self.play_station(station)

    def play_pause(self):
        self._take_control()
        if not self._is_playing():
            logger.debug(f"Sending play command")
            self._controller.play()
        else:
            logger.debug(f"Sending pause command")
            self._controller.pause()

    def prev_next(self, is_next: bool):
        station = self.current_station()
        if station is not None:
            logger.info(f"Current station is {station}.")
            index = self._stations.get_index(station)
            if index is not None:
                logger.info(f"Current station index is {index}.")
                if is_next:
                    to_play = self._stations.next_station(index)
                else:
                    to_play = self._stations.prev_station(index)
                self.play_station(to_play)
                return
            else:
                logger.error(f"No index for current station. It must not be a sonos favourite.")
        else:
            logger.info(f"Not currently playing a radio station, doing next/prev command.")
            self._take_control()
            if is_next:
                self._controller.next()
            else:
                self._controller.previous()

    def current_station(self) -> Optional[Station]:
        info = self._controller.get_current_media_info()
        if is_station_uri(info['uri']):
            return Station(title=info['channel'], uri=info['uri'])
        else:
            logger.info(f"Currently playing from channel {info['channel']}, uri {info['uri']} not a station.")

    def query_station(self):
        current = self.current_station()
        if current is not None:
            self.play_station(current)

    def to_group(self, i: str): # todo - weird using strings for indices even though it comes from json config
        self._take_control()
        current_slaves = {s.player_name for s in self._controller.group if s != self._controller}
        target_slaves = set(GROUP_CONFIG[self._controller.player_name][i])
        if target_slaves != current_slaves:
            logger.info(f"Grouping with target slaves {target_slaves}. (Current slaves are: {current_slaves}")
            released_slaves = current_slaves - target_slaves
            new_slaves = target_slaves - current_slaves
            for slave in released_slaves:
                ZONES[slave].unjoin()
            for slave in new_slaves:
                ZONES[slave].join(self._controller)
        else:
            logger.info(f"Target slaves {target_slaves} are equivalent to the existing slaves.")



import json
import logging
import os
import socket
import threading
from email.header import Header
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import quote, unquote

import requests

from socomote.settings import MP3_LIB

START_PORT = 9001
END_PORT = 9999

logger = logging.getLogger('tts_server')

class TTSRequestHandler(BaseHTTPRequestHandler):

    def do_GET(self):
        try:
            filename = unquote(self.path[1:])
            file = self._get_mp3_file(filename)
            with file.open('rb') as file:
                encoded_basename = Header(self.path).encode()
                stat = os.fstat(file.fileno())
                self.send_response(200)
                self.send_header('Content-Length', str(stat[6]))
                self.send_header(
                    'Content-Type',
                    'audio/mpeg')
                self.send_header(
                    'Content-Disposition',
                    'attachment; filename="%s"' % encoded_basename)
                self.end_headers()
                self.wfile.write(file.read())
        except socket.error as error:
            print('Connection is closed by peer')
            pass
        except IOError as error:
            self.send_response(500)
            print('I/O error: %s' % (str(error)))
        return

    def _get_mp3_file(self, filename):
        file = MP3_LIB / filename
        if not file.exists():
            text = filename.replace('mp3', '')
            gen = f"https://freetts.com/Home/PlayAudio?Language=en-GB&Voice=en-GB-Standard-C&TextMessage={quote(text)}&type=0"
            resp = requests.get(gen)
            data = json.loads(resp.text)
            mp3_id = data['id']
            download = f"https://freetts.com/audio/{mp3_id}"
            resp = requests.get(download)
            with file.open("wb") as f:
                f.write(resp.content)
        return file


class TTSServer(HTTPServer):

    def __init__(self):
        self.ip_addr = self._detect_ip_addr()
        self.port = self._find_free_port()
        super().__init__((self.ip_addr, self.port), TTSRequestHandler)
        self._thread = threading.Thread(target=self.serve_forever, daemon=True)

    def __enter__(self):
        super().__enter__()
        self._thread.start()
        return self

    def get_uri(self, text: str):
        return f"http://{self.ip_addr}:{self.port}/{quote(text)}.mp3"

    def _is_port_in_use(self, port):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(('localhost', port)) == 0

    def _find_free_port(self):
        for port in range(START_PORT, END_PORT):
            if not self._is_port_in_use(port):
                return port
        raise Exception('Cannot find a free port to use')

    def _detect_ip_addr(self, ):
        """Return the local ip-address"""
        # Rather hackish way to get the local ip-address, recipy from
        # https://stackoverflow.com/a/166589
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ip_address = s.getsockname()[0]
        s.close()
        return ip_address



import logging
from typing import Iterable

from getkey import getkey, keys

from socomote.action import *

logger = logging.getLogger(__name__)

class InputHandler:

    def actions(self) -> Iterable[Action]:
        digit_buffer = ''
        while True:
            char = getkey()
            logger.debug(f"Received {repr(char)}")
            input = None
            if char.isdigit():
                digit_buffer += char
                logger.debug(f"Added {repr(char)} to buffer. Buffer is now {repr(digit_buffer)}.")
            elif char in (keys.ENTER, 'm'):
                is_enter = char == keys.ENTER
                logger.debug(f"Received {'ENTER' if is_enter else 'MODE'}. Handling buffer.")
                if digit_buffer == '':
                    logger.debug(f"Empty buffer, nothing to do.")
                elif is_enter:
                    input = digit_buffer
                else:
                    input = digit_buffer + "m"
                digit_buffer = ''
            else:
                if digit_buffer != '':
                    logger.info("Non-enter or int received while buffer is non-empty, clearing buffer.")
                    digit_buffer = ''
                input = char
            if input is not None:
                try:
                    logger.info(f"Handling input {repr(input)}.")
                    yield self.handle_input(input)
                except:
                    pass
            else:
                logger.info("No valid input. Continuing.")

    def handle_input(self, inp: str) -> Action:
        if inp == keys.UP:
            return VolUp()
        elif inp == keys.DOWN:
            return VolDown()
        elif inp == keys.LEFT:
            return Previous()
        elif inp == keys.RIGHT:
            return Next()
        elif inp == 'p':
            return PlayPause()
        elif inp == 's':
            return ShuffleStation()
        elif inp == 'q':
            return Query()
        elif inp[0].isdigit():
            if inp[-1].isdigit():
                station_ix = int(inp)
                return SelectStation(station_ix)
            elif inp[-1] == 'm':
                return SelectGroup(inp[:-1])
        logger.error(f"Unrecognised input {repr(input)}")



import logging
import sys
from argparse import ArgumentParser

from socomote.receiver import Receiver
from socomote.settings import ZONES, CONFIG

parser = ArgumentParser()
parser.add_argument("--zone", "-z", help="Sonos zone to control (defaults to 'zone' in config.json).")
parser.add_argument("--debug", "-d", action="store_true", help="Run in DEBUG mode.")

def main(receiver: Receiver):
    with receiver as r:
        r.start()
    return 0

if __name__ == '__main__':
    args = parser.parse_args()
    logging_level = logging.DEBUG if args.debug else logging.INFO

    logging.getLogger('soco').setLevel(logging.ERROR)
    logging.getLogger('urllib3').setLevel(logging.ERROR)
    logging.getLogger('getkey').setLevel(logging.ERROR)

    logging.basicConfig(
        stream=sys.stdout, level=logging_level, format="[%(levelname)s] - %(name)s: %(message)s"
    )
    logger = logging.getLogger(__name__)
    zone_name = args.zone if args.zone is not None else CONFIG["zone"]
    logger.info(f"Starting socomote receiver for zone '{zone_name}'...")
    controller = ZONES[zone_name]
    receiver = Receiver(controller)

    sys.exit(main(receiver))



from setuptools import setup

setup(
    name='socomote',
    version='0.0.1',
    packages=['socomote'],
    install_requires=[
        "soco >= 0.23.2",
        "getkey >= 0.6.5"
    ],
    description="Application for remote control of Sonos systems",
    author="Theo Dickson",
    author_email="tmsdickson@gmail.com",
    license='MIT',
)